from PIL import Image

import fm_image_tagger.ollama_caption as Caption
import fm_image_tagger.tagger as Tagger


def get_meta(image_path):
    positive = ""
    negative = ""
    try:
        with Image.open(image_path) as img:
            meta_data = img.text
            if "parameters" in meta_data:
                prompt = meta_data["parameters"].split("\n")
                positive = prompt[0]
                negative = prompt[1]
    except Exception as e:
        print(f"An error occurred: {e}")
    negative = negative.replace("Negative prompt: ", "")
    return (positive, negative)


def display(image_path, meta_flag, threshold=0.35):
    if meta_flag:
        (positive, negative) = get_meta(image_path)
        print(f"Positive prompt:\n{positive}")
        print(f"Negative prompt:\n{negative}")
    else:
        tagger = Tagger.WD14Tagger(Tagger.MOAT_MODEL_DSV2_REPO)
        tags = tagger.image_tag(image_path, threshold=threshold)
        print(f"Tags:\n{tags}")


def caption(image_path, model, for_lora):
    text = Caption.image2text(image_path, model, for_lora)
    print(f"Caption:\n{text}")



import os


def rename_files(directory, suffix):
    try:
        # Ensure the provided directory exists
        if not os.path.isdir(directory):
            print(f"The directory {directory} does not exist.")
            return

        files = os.listdir(directory)
        files.sort()  # Ensure a consistent ordering

        for index, filename in enumerate(files):
            old_path = os.path.join(directory, filename)
            if os.path.isfile(old_path):
                name, ext = os.path.splitext(filename)
                new_filename = f"{suffix}_{index + 1:04}{ext}"
                new_path = os.path.join(directory, new_filename)
                os.rename(old_path, new_path)
                print(f"Renamed {old_path} to {new_path}")

    except Exception as e:
        print(f"An error occurred: {e}")



import os
import sys


def process_add(file_path, keyword):
    try:
        new_tags = []
        with open(file_path, "r", encoding="utf-8") as file:
            content = file.read()
            tags = content.split(",")
            for tag in tags:
                check_tag = tag.strip()
                new_tags.append(check_tag)
            new_tags.append(keyword)
        output_tags = ", ".join(new_tags).replace("\n", " ") + "\n"
        with open(file_path, "w+", encoding="utf-8") as file:
            file.write(output_tags)
    except Exception as e:
        print(f"Failed to process {file_path}: {e}")


def process_remove(file_path, keyword):
    try:
        new_tags = []
        with open(file_path, "r", encoding="utf-8") as file:
            content = file.read()
            tags = content.split(",")
            for tag in tags:
                check_tag = tag.strip()
                if not check_tag == keyword:
                    new_tags.append(check_tag)
        output_tags = ", ".join(new_tags).replace("\n", " ") + "\n"
        with open(file_path, "w+", encoding="utf-8") as file:
            file.write(output_tags)
    except Exception as e:
        print(f"Failed to process {file_path}: {e}")


def process_trigger(file_path, keyword):
    try:
        new_tags = [keyword]
        with open(file_path, "r", encoding="utf-8") as file:
            content = file.read()
            tags = content.split(",")
            for tag in tags:
                check_tag = tag.strip()
                if not check_tag == keyword:
                    new_tags.append(check_tag)
        output_tags = ", ".join(new_tags).replace("\n", " ") + "\n"
        with open(file_path, "w+", encoding="utf-8") as file:
            file.write(output_tags)
    except Exception as e:
        print(f"Failed to process {file_path}: {e}")


def add(dir, word, trigger=False):
    if not os.path.isdir(dir):
        print(f"The specified path is not a directory: {dir}")
        sys.exit(1)

    for root, dirs, files in os.walk(dir):
        for file in files:
            if file.endswith(".txt"):
                file_path = os.path.join(root, file)
                if trigger:
                    process_trigger(file_path, word)
                else:
                    process_add(file_path, word)


def remove(dir, word):
    if not os.path.isdir(dir):
        print(f"The specified path is not a directory: {dir}")
        sys.exit(1)

    for root, dirs, files in os.walk(dir):
        for file in files:
            if file.endswith(".txt"):
                file_path = os.path.join(root, file)
                process_remove(file_path, word)



import os
import sys


def read_tag(file_path, totals):
    try:
        with open(file_path, "r", encoding="utf-8") as file:
            content = file.read()
            print(f"Processing {file_path}:")
            tags = content.split(",")
            for tag in tags:
                check_tag = tag.strip()
                if check_tag in totals:
                    totals[check_tag] += 1
                else:
                    totals[check_tag] = 1  # Initialize new tag with count as zero
    except Exception as e:
        print(f"Failed to process {file_path}: {e}")


def total(dir):
    if not os.path.isdir(dir):
        print(f"The specified path is not a directory: {dir}")
        sys.exit(1)

    tags = {}
    for root, dirs, files in os.walk(dir):
        for file in files:
            if file.endswith(".txt"):
                file_path = os.path.join(root, file)
                read_tag(file_path, tags)
    sorted_tags = sorted(
        tags.items(), key=lambda x: -x[1]
    )  # Sorting by count in descending order (most common first).
    print(f"tags: {sorted_tags}")



# This code was created with reference to the following source code:
#   Predict.py in the SmilingWolf/wd-tagger repository of HaggingFace
#   https://huggingface.co/spaces/SmilingWolf/wd-tagger/tree/main
#
# The model used is from SmilingWolf.
# The necessary models will be downloaded at runtime.

import huggingface_hub
import numpy as np
import onnxruntime as rt
import pandas as pd
from PIL import Image

# Dataset v3 series of models:
SWINV2_MODEL_DSV3_REPO = "SmilingWolf/wd-swinv2-tagger-v3"
CONV_MODEL_DSV3_REPO = "SmilingWolf/wd-convnext-tagger-v3"
VIT_MODEL_DSV3_REPO = "SmilingWolf/wd-vit-tagger-v3"
VIT_LARGE_MODEL_DSV3_REPO = "SmilingWolf/wd-vit-large-tagger-v3"
EVA02_LARGE_MODEL_DSV3_REPO = "SmilingWolf/wd-eva02-large-tagger-v3"

# Dataset v2 series of models:
MOAT_MODEL_DSV2_REPO = "SmilingWolf/wd-v1-4-moat-tagger-v2"
SWIN_MODEL_DSV2_REPO = "SmilingWolf/wd-v1-4-swinv2-tagger-v2"
CONV_MODEL_DSV2_REPO = "SmilingWolf/wd-v1-4-convnext-tagger-v2"
CONV2_MODEL_DSV2_REPO = "SmilingWolf/wd-v1-4-convnextv2-tagger-v2"
VIT_MODEL_DSV2_REPO = "SmilingWolf/wd-v1-4-vit-tagger-v2"

# Files to download from the repos
MODEL_FILENAME = "model.onnx"
LABEL_FILENAME = "selected_tags.csv"


class WD14Tagger:
    def __init__(self, model_repo):
        self.model_target_size = None
        self.last_loaded_repo = None
        # model download
        csv_path = huggingface_hub.hf_hub_download(
            model_repo,
            LABEL_FILENAME,
        )
        model_path = huggingface_hub.hf_hub_download(
            model_repo,
            MODEL_FILENAME,
        )
        tags_df = pd.read_csv(csv_path)
        name_series = tags_df["name"]
        self.tag_names = name_series.tolist()
        self.general_indexes = list(np.where(tags_df["category"] == 0)[0])
        model = rt.InferenceSession(model_path)
        _, height, width, _ = model.get_inputs()[0].shape
        self.model_target_size = max(height, width)
        self.last_loaded_repo = model_repo
        self.model = model

    def prepare_image(self, image_path):
        image = Image.open(image_path)
        original_width, original_height = image.size
        # image resizing and padding to fit model input size (256x256 is a typical value for ResNet-like models, but it can be any other)
        aspect_ratio = min(
            self.model_target_size / original_width,
            self.model_target_size / original_height,
        )
        new_size = (
            int(original_width * aspect_ratio),
            int(original_height * aspect_ratio),
        )
        resized_image = image.resize(new_size, Image.LANCZOS)
        square_image = Image.new(
            "RGB", (self.model_target_size, self.model_target_size), (255, 255, 255)
        )
        paste_position = (
            (self.model_target_size - new_size[0]) // 2,
            (self.model_target_size - new_size[1]) // 2,
        )
        square_image.paste(resized_image, paste_position)
        # Convert to numpy array
        image_array = np.asarray(square_image, dtype=np.float32)
        # Convert PIL-native RGB to BGR
        image_array = image_array[:, :, ::-1]
        return np.expand_dims(image_array, axis=0)

    def tagging(self, input, threshold=0.35):
        # run model
        input_name = self.model.get_inputs()[0].name
        label_name = self.model.get_outputs()[0].name
        preds = self.model.run([label_name], {input_name: input})[0]
        # collect labels
        labels = list(zip(self.tag_names, preds[0].astype(float)))
        general_names = [labels[i] for i in self.general_indexes]
        general_res = [x for x in general_names if x[1] > threshold]
        general_res = dict(general_res)
        sorted_general_strings = sorted(
            general_res.items(),
            key=lambda x: x[1],
            reverse=True,
        )
        #
        sorted_general_strings = [x[0] for x in sorted_general_strings]
        sorted_general_strings = (
            ", ".join(sorted_general_strings).replace("(", "\(").replace(")", "\)")
        )
        return sorted_general_strings

    def image_tag(self, image_path, threshold=0.35):
        image = self.prepare_image(image_path)
        return self.tagging(image, threshold)



import os
import sys

import fm_image_tagger.ollama_caption as Caption
import fm_image_tagger.tagger as Tagger


def createTagFiles(path, threshold=0.35):
    if not os.path.isdir(path):
        print(f"The specified path is not a directory: {path}")
        sys.exit(1)

    tagger = Tagger.WD14Tagger(Tagger.SWINV2_MODEL_DSV3_REPO)
    for root, dirs, files in os.walk(path):
        for file in files:
            if file.endswith(".png") or file.endswith(".jpg") or file.endswith(".jpeg"):
                file_path = os.path.join(root, file)
                tags = tagger.image_tag(file_path, threshold=threshold)
                # create tag file
                base = os.path.splitext(file_path)[0]
                with open(f"{base}.txt", "w", encoding="utf-8") as f:
                    f.write(tags)
    print("Done.")


def createCaptionFiles(path, model, for_lora):
    if not os.path.isdir(path):
        print(f"The specified path is not a directory: {path}")
        sys.exit(1)

    for root, dirs, files in os.walk(path):
        for file in files:
            if file.endswith(".png") or file.endswith(".jpg") or file.endswith(".jpeg"):
                file_path = os.path.join(root, file)
                caption = Caption.image2text(file_path, model, for_lora)
                # create tag file
                base = os.path.splitext(file_path)[0]
                with open(f"{base}.txt", "w", encoding="utf-8") as f:
                    f.write(caption)
    print("Done.")



import ollama

PROMPT_FOR_LORA = """
Please explain the input image. Describe as many detailed elements as possible. 
Absolutely do not output or mention the following elements:
- Eyes color
- Features reminiscent of an animal
- Has a tail (and a description of its appearance)
- Description of the ears (e.g., "with pointed ears like a fox", "animal traits such as large ears")
- Hair color
- Hairstyle
- Yellow streaked hair and ahoge (cowlick)

Absolutely add output or mention the following elements:
- 3dcg
- polygon
- vr avatar
- not furry
"""

PROMPT_SIMPLE = """
Please explain the input image. Describe as many detailed elements as possible. 
"""

CANCELED = "sorry"


def image2text(image, model, for_lora=False):
    if for_lora:
        prompt = PROMPT_FOR_LORA
    else:
        prompt = PROMPT_SIMPLE
    response = ollama.chat(
        model=model,
        messages=[{"role": "user", "content": prompt, "images": [image]}],
    )
    #
    if CANCELED in response["message"]["content"]:
        print(f"Analysis rejected: {image}")
    return response["message"]["content"]



import click

import fm_image_tagger.command_tagger as command_tagger
import fm_image_tagger.display as Display
import fm_image_tagger.rename as Rename
import fm_image_tagger.tag_edit as Editor
import fm_image_tagger.total as Total


# CLI main function
@click.group()
def cli():
    pass


@cli.command(help="Create tag files")
@click.argument("path", type=str)
@click.option(
    "-t", "--threshold", default=0.35, help="score threshold (default:0.35)", type=float
)
def tagger(path, threshold):
    command_tagger.createTagFiles(path, threshold)


@cli.command(help="Create caption files, use Ollama vision-model")
@click.argument("ollama-vision-model", type=str)
@click.argument("path", type=str)
@click.option("-l", "--for-lora", is_flag=True, help="LoRA caption mode")
def caption(ollama_vision_model, path, for_lora):
    command_tagger.createCaptionFiles(path, ollama_vision_model, for_lora)


@cli.command(help="Rename with the serial number")
@click.argument("suffix", type=str)
@click.argument("path", type=str)
def rename(suffix, path):
    Rename.rename_files(path, suffix)


@cli.command(help="add word to tag files")
@click.argument("keyword", type=str)
@click.argument("path", type=str)
@click.option("-f", "--first", is_flag=True, help="Add to tags top")
def add(keyword, path, first):
    Editor.add(path, keyword, first)


@cli.command(help="remove word from tag files")
@click.argument("keyword", type=str)
@click.argument("path", type=str)
def remove(keyword, path):
    Editor.remove(path, keyword)


@cli.command(help="remove word from tag files")
@click.argument("path", type=str)
def total(path):
    Total.total(path)


@cli.command(help="display picture information")
@click.argument("path", type=str)
@click.option("-m", "--meta", is_flag=True, help="display prompt in metainfo")
@click.option(
    "-t", "--threshold", default=0.35, help="score threshold (default:0.35)", type=float
)
@click.option(
    "-c",
    "--caption-model",
    default="",
    help="Caption display when Ollama model is specified",
    type=str,
)
@click.option("-l", "--for-lora", is_flag=True, help="LoRA caption mode")
def display(path, meta, threshold, caption_model, for_lora):
    if len(caption_model) > 0:
        Display.caption(path, caption_model, for_lora)
    else:
        Display.display(path, meta, threshold=threshold)


# cli app entry point
def run():
    cli()
